import asyncore, socket
from Queue import Queue
import re
import random

class ClientConnection(object, asyncore.dispatcher):

    def __init__(self):
        asyncore.dispatcher.__init__(self)
        self.create_socket(socket.AF_INET, socket.SOCK_STREAM)

        self.verified = False
        self.authenticated = False
        self.authentification_sent = False

        self.retrieved_user_list = False
        self.request_user_list_sent = False

        # Generate a random user name
        self.user = 'Anonymous' + str(int(random.random()*10000))

        # Connection/chat event callbacks
        self.event_list = { }

        self.send_queue = Queue()
        self.send_buffer = ''

    def init_loop(self):
        """Initialise the asyncore loop (blocking)."""
        # Set a timeout of 0.1 seconds and use poll() when available.
        asyncore.loop(.1, True)

    def debug_log(self, msg):
        if 'debug' in self.event_list:
            self.event_list['debug'](self, msg)
        else:
            print msg

    def connect(self, addr):
        self.host = addr[0]
        self.port = addr[1]
        super(ClientConnection, self).connect(addr)

    def send_raw(self, cmd):
        """
        Send raw command to the chat server.
        """
        self.send_queue.put(cmd)

    def retrieve_users(self):
        """
        Retreive list of users active at the chat server.
        """

        self.request_user_list_sent = True
        self.retrieved_user_list = False
        self.send_queue.put('NAMES')

    def username(self, name=''):
        if not name:
            return self.user
        self.user = name
        self.authenticated = False
        self.authentification_sent = True
        self.send_queue.put('USER %s' % name)

    def handle_connect(self):
        """
        Called when a connection is established.
        """

        if 'connect' in self.event_list:
            self.event_list['connect'](self)

    def handle_close(self):
        self.close()

    def handle_read(self):
        buf = ''
        multi_line = False
        response_format = False

        # Receive a (multiline) message.
        while True:
            chunk = self.recv(1)
            if not chunk:
                raise RuntimeError('socket connection broken')
            if chunk in ['-','+']:
                response_format = True
            elif chunk == ':':
                multi_line = True
            elif chunk == '\n' and buf[-1] == '\r':
                if not multi_line:
                    break
                if buf[-3:] == '\r\n\r':
                    # Remove trailing '\n\r' from multi line response
                    buf = buf[:-2]
                    break
            buf += chunk

        buf = buf[:-1]
        self.debug_log('< %s' % buf)

        # Invoke the proper callback function.
        if not self.verified:
            self.verify_server(buf)
        elif response_format:
            self.parse_response(buf)
        else:
            self.parse_notification(buf)

    def handle_write(self):
        if not self.send_buffer:
            self.send_buffer = self.send_queue.get()
            self.debug_log('> %s' % self.send_buffer)
            self.send_buffer += '\r\n'
        sent = self.send(self.send_buffer)
        self.send_buffer = self.send_buffer[sent:]

    def verify_server(self, buf):
        match = re.match('^CHAT/(\d\.\d)/([^\x00-\x1F/:]+)$', buf)

        if not match:
            self.debug_log('Failed to verify connection')
            sys.exit(-1)

        self.debug_log('Server version is %s' % match.group(1))

        if 'verify' in self.event_list:
            self.event_list['verify'](self, buf, match.group(1))

        self.verified = True
        self.send_queue.put('CHAT')

    def parse_response(self, buf):
        if 'response' in self.event_list:
            self.event_list['response'](self, buf)

        # First authenticate the user.
        if not self.authenticated:
            if not self.authentification_sent:
                self.send_queue.put('USER %s' % self.user)
                self.authentification_sent = True
            elif buf[1:9] == 'Username':
                # TODO: handle 'username is taken'.
                self.authenticated = True
                if 'authenticated' in self.event_list:
                    self.event_list['authenticated'](self, self.user)

        if not self.authenticated:
            return

        # After authentification, fetch the list of users currently logged in.
        if not self.retrieved_user_list:
            if not self.request_user_list_sent:
                self.request_user_list_sent = True
                self.retrieve_users()
            else:
                # TODO: process user list
                self.retrieved_user_list = True
                msg = 'Users: [%s]' % '] ['.join(buf.split('\r\n')[1:])
                if 'info' in self.event_list:
                    self.event_list['info'](self, msg)
                else:
                    self.debug_log(msg)

    def parse_notification(self, buf):
        match = re.match('^SAY ([^\r\n:/]+)/(.+)', buf)
        if match:
            if 'message' in self.event_list:
                self.event_list['message'](self, match.group(1), match.group(2))
        elif 'notify' in self.event_list:
            self.event_list['notify'](self, buf)

    def writable(self):
        write = not self.send_queue.empty() or len(self.send_buffer)
        return write

if __name__ == '__main__':
    client = ClientConnection()
    client.connect(('ow150.science.uva.nl', 16897))
    client.init_loop()
